/* cg.cc */
#include <iostream>
#include <fstream>
#include <assert.h>

#include "cg.hh"
#include "vcg.hh"
#include "opts.hh"
#include "eopt.hh"
#include "ls.hh"
#include "verbose.hh"
#include "ilpd.hh"
#include "gaopt.hh"

bool CG::b_initialized = false;
char *CG::archi = NULL;
Emitter *CG::emitter = NULL;
Architecture *CG::architecture = NULL;
CG *gp_cg = NULL;
TimeProfile *CG::empty_time_profile = NULL;
Flags CG::flags;

#ifdef __cplusplus
extern "C" {
#endif

  void cg_initialize(int argc, char *argv[]) { CG::initialize(argc, argv); }
  void cg_finalize(void) { CG::finalize(); }
  void cg_function(lcc::Symbol f, lcc::Symbol caller[], lcc::Symbol callee[], int ncalls) {
    gp_cg->function(f, caller, callee, ncalls);
  }
  lcc::Node cg_gen(lcc::Node p) {
    assert(gp_cg);
    return gp_cg->gen(p);
  }
  void cg_emit(lcc::Node p, char *archi) { 
    assert(gp_cg);
    gp_cg->emit(p, archi);
  }

#ifdef __cplusplus
}
#endif

CG::CG()
{
  //pu = NULL;
}

CG::~CG()
{
  std::list< std::pair< Schedule *, DAG * > >::iterator it= schedule.begin();
  for(; it!=schedule.end(); ++it){
    delete (*it).first;
  }
}

extern int ARM_ASSEMBLER_MODE; /* optcg/src/opt.c */
extern int OPT_C_DEBUG; /* optcg/src/opt.c */

void CG::initialize(int argc, char *argv[])
{
  gp_cg = new CG(); //global.
  int i_param;

  for(i_param = 1; i_param < argc; i_param++) {
    if(argv[i_param][0] != '-') {
      Flags::program_file = argv[i_param];
      break;
    }

    if(!strcmp(argv[i_param], "-?") || 
       !strcmp(argv[i_param], "-help") ||
       !strcmp(argv[i_param], "--help")) {
      flags.usage();
      exit(0);
    }
    
    if(!strcmp(argv[i_param], "-emitter:mnemonic")){
      Emitter::always_mnemonic = true;
    }
    if(!strcmp(argv[i_param], "-emitter:debug")){
      Emitter::debug_mode = true;
    }
    if(!strcmp(argv[i_param], "-arm_asm"))
    {
      ARM_ASSEMBLER_MODE = 1;
    }
    if(!strcmp(argv[i_param], "-opt_c_debug")){
      OPT_C_DEBUG = 1;
    }
    
    if(!strncmp(argv[i_param], "-archi=", 7)) {
      archi = &argv[i_param][7];
      //Check if the file exists
      FILE *f = fopen(archi, "r");
      if(!f) {
	std::cerr << "\nFATAL ERROR\n  File: " 
		  << archi << " does not exist or is unreadable.\n\n";
	exit(-1);
      }
      fclose(f);
    }

    if(!strncmp(argv[i_param], "-vcg", 4)) {
      flags.b_vcg = true;
    }
    if(!strncmp(argv[i_param], "-ilp", 4)) {
      flags.b_ilp = true;
    }

    if(!strncmp(argv[i_param], "-ddep", 5)) {
      flags.no_ddep_link = true;
    }

    if(!strncmp(argv[i_param], "-dot", 4)) {
      flags.b_dot = true;
    }

    if(!strncmp(argv[i_param], "-verb=", 6)) {
      int verb_level = atoi((char *)&argv[i_param][6]);
      Verbose::verbosityLevel = atoi((char *)&argv[i_param][6]);
    }

    if(!strncmp(argv[i_param], "-opt=", 5)) {
      if(!strncasecmp(&(argv[i_param][5]), "e", 1)) {
	flags.opt_type = Flags::ENERGY;
      }
      if(!strncasecmp(&(argv[i_param][5]), "t", 1)) {
	flags.opt_type = Flags::TIME;
      }
      if(!strncasecmp(&(argv[i_param][5]), "g", 1)) {
        flags.opt_type = Flags::GA;
      }
    }

    if(!strncmp(argv[i_param], "-bb=", 4)) {
      flags.i_bb = atoi((char *)&argv[i_param][4]);
    }

    if(!strncmp(argv[i_param], "-hs=", 4)) {
      flags.i_hs = atoi((char *)&argv[i_param][4]);
    }

    if(!strncmp(argv[i_param], "-ts=", 4)) {
      flags.i_ts = atoi((char *)&argv[i_param][4]);
    }
    

    // Genetic Algorithm flags
    if(!strncmp(argv[i_param], "-gapthreads", 12)) {
      flags.b_ga_pthreads = true;
    }

    if(!strncmp(argv[i_param], "-gamaxgen=", 10)) {
      flags.i_ga_max_gen = atoi((char *)&argv[i_param][10]);
    }
    
    if(!strncmp(argv[i_param], "-gasize=", 8)) {
      flags.i_ga_size = atoi((char *)&argv[i_param][8]);
    }
    
    if(!strncmp(argv[i_param], "-gastable_stop=", 15)) {
      flags.i_ga_stable_stop = atoi((char *)&argv[i_param][15]);
    }
    
    if(!strncmp(argv[i_param], "-gamutate_node=", 15)) {
      flags.d_ga_mutate_node = atof((char *)&argv[i_param][15]);
    }
    
    if(!strncmp(argv[i_param], "-gamutate_inst=", 15)) {
      flags.d_ga_mutate_inst = atof((char *)&argv[i_param][15]);
    }
    
    if(!strncmp(argv[i_param], "-gamutate_transfer=", 19)) {
      flags.d_ga_mutate_transfer = atof((char *)&argv[i_param][19]);
    }
    
    if(!strncmp(argv[i_param], "-gaparents_survive", 18)) {
      flags.b_ga_parents_survive = true;
    }

    if(!strncmp(argv[i_param], "-ganoparents_survive", 20)) {
      flags.b_ga_parents_survive = false;
    }
    
    if(!strncmp(argv[i_param], "-ganew_children=", 16)) {
      flags.d_ga_new_children = atof((char *)&argv[i_param][16]);
    }
    
    if(!strncmp(argv[i_param], "-gamultiple", 11)) {
      flags.b_ga_allow_multiple_mutations = true;
    }

    if(!strncmp(argv[i_param], "-ganomultiple", 13)) {
      flags.b_ga_allow_multiple_mutations = false;
    }
    
    if(!strncmp(argv[i_param], "-gasurvivor_truncation", 22)) {
      flags.b_ga_survivor_truncation = true;
      flags.b_ga_survivor_roulette = false;
    }
    
    if(!strncmp(argv[i_param], "-gasurvivor_roulette", 20)) {
      flags.b_ga_survivor_roulette = true;
      flags.b_ga_survivor_truncation = false;
    }

    if(!strncmp(argv[i_param], "-gabench", 8)) {
      flags.b_ga_bench = true;
    }
    
    if(!strncmp(argv[i_param], "-garandom_search", 16)) {
      flags.b_ga_random_search = true;
    }
  }

  if(!archi) {
    std::cerr << "\nFATAL ERROR:\n"
      "  Target architecture specification in ADML is required.\n"
      "  Option to lcc: -Wf-archi=<XML file>\n\n";
    exit(-2);
  }
  architecture = new Architecture(archi);
  assert(architecture);
  architecture->parse();
  emitter = new Emitter(architecture);
  
  Flags::architecture_file = archi;
  b_initialized = true;
  
}

void CG::finalize(void)
{
  delete gp_cg;
  delete empty_time_profile;
  delete architecture;
  delete emitter;
  //delete solution_space;
  b_initialized = false;
}

lcc::Node CG::gen(lcc::Node p)
{
  VERB3 << "CG::gen (DAG) - " << lcc::opname(p->op) << "\n";
  //Build forest of DAGs
  PU pu = PU(p, architecture, flags);

  //debug
  if(flags.b_vcg || flags.b_dot) {
    pu.print(flags.b_vcg, flags.b_dot);
  }

  //Generate ILP
  if(flags.b_ilp) {

    char file_name[12];
    static int dag_number;
    std::vector< DAG * > dags = pu.get_dags();
    std::vector< DAG * >::iterator d_iter = dags.begin();
    for(; d_iter != dags.end(); ++d_iter) {
      DAG *dag = *d_iter;

      sprintf(file_name, "%s%.2d%s", "ilp", dag_number, ".dat");
      std::ofstream dat_file(file_name);
      dat_file << "#rcc -archi=" << Flags::architecture_file
	       << " -bb=" << Flags::i_bb
	       << " " << Flags::program_file
	       << std::endl;
      dat_file << "data;\n" << std::endl;
      ILPD(*dag, dat_file);
      architecture->ilp(dat_file);
      dat_file.close();

      dag_number++;
    }
  }

  Optimizer *opt;

  switch(flags.opt_type) {
  case Flags::ENERGY: {
    opt = new EOptimist(architecture);
    break;
  }
  case Flags::TIME: {
    opt = new Optimist(architecture);
    break;
  }
  case Flags::GA: {
    opt = new GAOptimist(architecture);
    break;
  }
  default:
    opt = new Optimist(architecture);
  }

  std::vector< DAG * >::iterator dag_iter = pu.get_dags().begin();
  //flags.i_bb has default value -1. Has it been given a new value?
  bool b_single_bb = (flags.i_bb != -1) ? true : false;

  for(; dag_iter != pu.get_dags().end(); ++dag_iter) {
    //select bb (mater note: arithmetics with a flag is ugly)
    if(b_single_bb && flags.i_bb != 0) {
      flags.i_bb--;
      continue;
    }
    //Now we are at a DAG we wish to schedule
    DAG *bb = *dag_iter;
    Schedule *s = opt->schedule(*bb, flags);
    schedule.push_back(std::make_pair(s, bb));
    if(b_single_bb) {
      flags.i_bb = -2;
    }
  }

  delete opt;
  return p; 
}

void CG::function(lcc::Symbol f, 
		  lcc::Symbol *caller,
		  lcc::Symbol *callee,
		  int ncalls)
{
  VERB3 << "#Function\n_" << f->name << "\n";
}

void CG::emit(lcc::Node p, char *archi)
{
  VERB3 << "CG::emit - " << lcc::opname(p->op) << "\n";

  std::list< std::pair< Schedule *, DAG * > >::iterator si;
  for(si = schedule.begin(); si != schedule.end(); ++si) {
    //emit code Anders method class Emitter();
    emitter->emit(si->first);
  }
  std::list< std::pair< Schedule *, DAG * > >::iterator it= schedule.begin();
  for(; it!=schedule.end(); ++it){
    delete (*it).first;
  }
  schedule.clear();
}

Emitter *CG::get_emitter()
{
  return emitter;
}
