#include "gagnode.hh"
#include "gautility.hh"

namespace GA
{


GNode::GNode(const GNode& _gnode)
{
  copy_data(_gnode);
}

/**
 * Will NOT delete the TargetInstruction when the object is
 * destroyed, but will delete its TimeProfile and SpaceProfile.
 */
GNode::~GNode()
{
  delete_me();
}

inline void GNode::delete_me()
{
  if (time_profile != 0) {
    delete time_profile;
    time_profile = 0;
  }
#ifdef _SP_
  if (space_profile != 0) {
    delete space_profile;
    space_profile = 0;
  }
#endif //_SP_
}

const GNode& GNode::operator=(const GNode& _gnode)
{
  if (&_gnode != this) {
    delete_me();
    copy_data(_gnode);
  }
  return *this;
}

void GNode::copy_data(const GNode& _gnode)
{
  reference_time = _gnode.reference_time;
  tau = _gnode.tau;
  z = _gnode.get_z();
  z_choice = _gnode.z_choice;
  target_instruction = _gnode.target_instruction;
  
  if (_gnode.time_profile == 0) {
    time_profile = 0;
  } else {
    time_profile = new TimeProfile(*(_gnode.time_profile));
  }
#ifdef _SP_
  if (_gnode.space_profile == 0) {
    space_profile = 0;
  } else {
    space_profile = new SpaceProfile(*(_gnode.space_profile));
  }
#endif //_SP_
  
  type = _gnode.type;
}


};

std::ostream& operator<<(std::ostream & _out, const GA::GNode& g)
{
  _out << "ref: " << g.get_reference_time() 
       << ", tau: " << g.get_tau()
       << ", choice: "
       << ((g.get_z_choice() != 
	   boost::graph_traits<DAG>::null_vertex()) ? 
	   g.get_z_choice() : 404)
       << ", tinst: ";
  if (g.get_target_instruction() != NULL){
    _out << *(g.get_target_instruction());
  }
  _out << ", z: " << GA::print_zset(g.get_z())
       << ", tp: " << *(g.get_time_profile())
    //<< ", sp: " << *(g.get_space_profile())
       << std::endl;

  return _out;
}

