#include "gautility.hh"

#include <iostream>
#include <iomanip>
#include <sstream>
#include "tp.hh"
#include "verbose.hh"


namespace GA
{


/**
 * Returns an integer value in the range of [s1, s2[ .
 * TODO: Inline?
 */
int random_interval(int s1, int s2)
{
  return s1 + int(((double) (s2-s1))*rand() / (RAND_MAX+1.0));
}

double drandom_interval(double s1, double s2)
{
  return s1 + ((s2-s1)*rand() / (RAND_MAX+1.0));
}

/**
 * Returns a double value in the range of [0.0, 1.0[
 * TODO: Inline?
 */
double drandom()
{
  return double(rand()) / double(RAND_MAX+1.0);
}

/**
 * Returns true or false.
 * TODO: Inline?
 */
bool random_boolean()
{
  return int(2.0*rand() / (RAND_MAX+1.0));
}

std::string print_compare_gnodes(GNode& g1, GNode& g2)
{
  std::stringstream all;
  
  TimeProfile* tp1 = g1.get_time_profile();
  TimeProfile* tp2 = g2.get_time_profile();
  
  if (tp1 == 0 || tp2 == 0) {
    std::cout << "One of the TimeProfiles is null.\n";
    return "";
  }
  
  std::stringstream out;
  all.setf(std::ios::left, std::ios::adjustfield);
  
  out.str("");
  out << "z: {";
  vertex_set_t::const_iterator vit = g1.get_z().begin();
  for (; vit != g1.get_z().end();) {
    out << *vit;
    if (++vit != g1.get_z().end()) {
      out << ", ";
    }
  }
  out << "}";
  all << std::setw(40) << out.str();
  out.str("");
  out << "z: {";
  vit = g2.get_z().begin();
  for (; vit != g2.get_z().end();) {
    out << *vit;
    if (++vit != g2.get_z().end()) {
      out << ", ";
    }
  }
  out << "}";
  all << out.str() << std::endl;
  
  
  out.str("");
  out << "nodes: { ";
  std::copy(tp1->get_nodes().begin(), tp1->get_nodes().end(), 
      std::ostream_iterator< vertex_id >(out, " "));
  out  << "}";
  all << std::setw(40) << out.str();
  out.str("");
  out << "nodes: { ";
  std::copy(tp2->get_nodes().begin(), tp2->get_nodes().end(), 
            std::ostream_iterator< vertex_id >(out, " "));
  out << "}";
  all << out.str() << std::endl;
  
  out.str("");
  out << "latencies: (v, l): {";
  std::map< vertex_id, int >& latencies1 = tp1->get_latency_map();
  std::map< vertex_id, int >::iterator li = latencies1.begin(),
    li_end = latencies1.end();
  for(; li != li_end; ++li) {
    out << "(" << (*li).first << ", " << (*li).second << ") ";
  }
  out << "}";
  all << std::setw(40) << out.str();
  out.str("");
  out << "latencies: (v, l): {";
  std::map< vertex_id, int >& latencies2 = tp2->get_latency_map();
  li = latencies2.begin();
  li_end = latencies2.end();
  for(; li != li_end; ++li) {
    out << "(" << (*li).first << ", " << (*li).second << ") ";
  }
  out << "}";
  all << out.str() << std::endl;
  
  
  
  int f1 = (*tp1).size();
  int f2 = (*tp2).size();
  all << std::setw(40) << "time profile:" << "time profile:" << std::endl;
  if (f1 == f2) {
    for(int i = 0; i < f1; i++) {
      out.str(""); 
      int size = (*tp1).get_rtab()[i].size();
      out << i << " (" << size << "): ";
      for(int j = 0; j < size; j++) {
        out << (*tp1).get_rtab()[i][j] << " ";
      }
      all << std::setw(40) << out.str();
      out.str("");
      size = (*tp2).get_rtab()[i].size();
      out << i << " (" << size << "): ";
      for(int j = 0; j < size; j++) {
        out << (*tp2).get_rtab()[i][j] << " ";
      }
      all << out.str() << std::endl;
    }
  } else {
    std::cerr << "Got TimeProfiles with different length. Will not print." << std::endl;
  }
  
  // TODO: Printing for SpaceProfile.
  
  return all.str();
}

std::string print_zset(const vertex_set_t& z)
{
  std::stringstream all;
  
  all << "{";
  vertex_set_t::const_iterator vit = z.begin();
  for (; vit != z.end();) {
    all << *vit;
    if (++vit != z.end()) {
      all << ", ";
    }
  }
  all << "}";
  
  return all.str();
}

std::string print_zlist(const std::list<vertex_id>& z)
{ 
  std::stringstream all;
  
  all << "{";
  std::list<vertex_id>::const_iterator vit = z.begin();
  for (; vit != z.end();) {
    all << *vit;
    if (++vit != z.end()) {
      all << ", ";
    }
  }
  all << "}";
  
  return all.str();
}


};
