#include "matcher.hh"
#include <algorithm>
#include <assert.h>
#include "verbose.hh"

Matcher::Matcher(Architecture *_architecture)
  : architecture(_architecture)
{
}

Matcher::Matcher(const Matcher & _m)
  : architecture(_m.architecture),
    match_map(_m.match_map)
{
}

Matcher::~Matcher()
{
}

std::pair< Matcher::match_map_iter, Matcher::match_map_iter >
Matcher::matches(vertex_id & _u_id, Node & _current_node, DAG & _dag)
{
  //build the match map
  match_map.clear();
  std::vector< pvertex_id > starting_nodes;
  //LCC IR node value
  lcc::Node u = get_lcc_node(_u_id, _dag);
  std::pair< match_iter, match_iter > match_candidates = architecture->match_start(u->op);
  //  < pattern_t *, TargetInstruction * > *

  for( ; match_candidates.first != match_candidates.second; ++(match_candidates.first)) {
    pattern_t *pattern = (*(*(match_candidates.first))).first;
    TargetInstruction *target_instruction = (*(*(match_candidates.first))).second;
    //pvertex: pattern vertex (see pattern.hh)
    pvertex_iter vi, vi_end;
    starting_nodes.clear();
    vertex_kind_map::type vkm = vertex_kind_property(*pattern);
    vertex_info_map::type vim = vertex_info_property(*pattern);

    //starting nodes of the pattern (to be constructed when building/parsing) FIXME
    pvertex_id start_node;
    bool b_start_node = false;
    for(tie(vi, vi_end) = vertices(*pattern); vi != vi_end; ++vi) {
      if((vkm[*vi] == OPNODE) && vim[*vi].indeg == 0) {
	if(!b_start_node) {
	  std::list< int > & ops = vim[*vi].ops;
	  std::list< int >::iterator oi;
	  for(oi = ops.begin(); oi != ops.end(); ++oi) {
	    if(*oi == u->op) {
	      start_node = *vi;
	      b_start_node = true;
	      break;
	    }
	  }
	}
	if(start_node != *vi || !b_start_node) {
	  starting_nodes.push_back(*vi);
	}
      }
    } 
    assert(b_start_node);

    if(starting_nodes.size() == 0) {
      std::vector< iso_map_s > iso_maps; //(1);
      if(check(start_node, *pattern, _u_id, _dag, _current_node, 
	       *target_instruction, iso_maps)) {

	//Add starting nodes as isomorph
	for(int i = 0; i < iso_maps.size(); i++) {
	  iso_maps[i].iso_map[start_node] = _u_id;
	}

	match_map[target_instruction] = make_pair(pattern, iso_maps);

      }    
    } else {
      //All permutation on remining nodes
      NodeAttr *node_attr = _current_node.get_node_attr();
      std::set< int > ops; //All op for the leave pattern
      for(int i = 0; i < starting_nodes.size(); i++) {
	std::list< int > & tmp_ops = vim[starting_nodes[i]].ops;
	for(std::list< int >::iterator v_iter = tmp_ops.begin(); 
	    v_iter != tmp_ops.end(); ++v_iter) {
	  ops.insert(*v_iter);
	}
      }
      std::vector< vertex_id > x_i;

      //for all nodes from z\u_id
      for(vertex_set_t::iterator z_iter = node_attr->z.begin(); z_iter != node_attr->z.end();
	  ++z_iter) {
	if(*z_iter == _u_id) {
	  continue;
	}
	int op = get_lcc_node(*z_iter, _dag)->op;
	if(ops.find(op) != ops.end()) {
	  x_i.push_back(*z_iter);
	}
      }
      int start_size = starting_nodes.size();
      if(x_i.size() < start_size) {
	//failed!
	continue;
      }
      //Check all permutations
      std::sort(x_i.begin(), x_i.end());

      std::vector< iso_map_s > base_iso_maps;
      base_iso_maps.clear();

      if(!check(start_node, *pattern, _u_id, _dag, 
		_current_node, *target_instruction, base_iso_maps)) {
	continue; //for loop
      }

      //Add initial nodes
      for(int i = 0; i < base_iso_maps.size(); i++) {
	base_iso_maps[i].iso_map[start_node] = _u_id;
      }

      std::vector< iso_map_s > iso_maps;
      std::vector< std::vector< iso_map_s > >partial_iso_maps(start_size);

      do {
	bool b_success = false;
	for(int i = 0; i < start_size; i++) {
	  partial_iso_maps[i].clear();
	  if(!check(starting_nodes[i], *pattern, x_i[i], _dag, 
		   _current_node, *target_instruction, partial_iso_maps[i])) {
	    break; //for loop
	  }
	  //compose partial_iso_map
	  b_success = true;
	}

	if(b_success) {
	  //Concatenate base_iso_maps \bowtie parital_iso_maps
	  for(int i = 0; i < base_iso_maps.size(); i ++) {
	    //Check for same root!
	    vertex_id root = base_iso_maps[i].root;
	    for(int j = 0; j < start_size; j++) {
	      for(int jj = 0; jj < partial_iso_maps[j].size(); jj++) {

		if(partial_iso_maps[j][jj].root != root) {
		  //different roots (iso not equal)
		  continue;
		}
		//Same root
		iso_maps.push_back(iso_map_s());
		iso_map_s & current_iso = iso_maps.back();
		//Nodes
		for(node_map_t::iterator k = partial_iso_maps[j][jj].iso_map.begin(), 
		      k_end =  partial_iso_maps[j][jj].iso_map.end(); k != k_end; ++k) {
		  current_iso.iso_map[ k->first ] = k->second;
		}
		//Arguments
		for(node_map_t::iterator k = partial_iso_maps[j][jj].arg_map.begin(), 
		      k_end =  partial_iso_maps[j][jj].arg_map.end(); k != k_end; ++k) {
		  current_iso.arg_map[ k->first ] = k->second;
		}
		//For base nodes (start_node <-> u_id)
		//Nodes
		for(node_map_t::iterator k = base_iso_maps[i].iso_map.begin(),
		      k_end = base_iso_maps[i].iso_map.end(); k != k_end; ++k) {
		  current_iso.iso_map[ k->first ] = k->second;
		}
		//Arguments
		for(node_map_t::iterator k = base_iso_maps[i].arg_map.begin(),
		      k_end = base_iso_maps[i].arg_map.end(); k != k_end; ++k) {
		  current_iso.arg_map[ k->first ] = k->second;
		}
		current_iso.root = root;

		//Matching of current permutation
		for(int m = 0; m < start_size; m++) {
		  current_iso.iso_map[ starting_nodes[m] ] = x_i[m];
		}
	      }
	    }
	  }
	}
      } while(std::next_permutation(x_i.begin(), x_i.end()));

      match_map[target_instruction] = make_pair(pattern, iso_maps);

    }
  }
  //  match_map_t::iterator mi = match_map.begin(), mi_end = match_map.end();
  match_map_iter mi = match_map.begin(), mi_end = match_map.end();
  
  //debug
  VERB3 << " match_map.size: " << match_map.size() << "\n";
  if(Verbose::verbosityLevel > 1) {
    VERB3 << "\nmatches:\n";
    for( ; mi != mi_end; ++mi) {
      VERB3 << "***\nTargetInstruction " << *(mi->first) << "\n";
      for(std::vector< iso_map_s >::iterator iso_iter = (mi->second.second).begin();
	  iso_iter != (mi->second.second).end(); ++iso_iter) {
	VERB3 << "###\nArguments (p->v): ";
	for(node_map_t::iterator arg_iter = iso_iter->arg_map.begin(); 
	    arg_iter != iso_iter->arg_map.end(); ++arg_iter) {
	  VERB3 << arg_iter->first << "->" << arg_iter->second << " ";
	}
	VERB3 << "\nIso (p->v): ";
	for(node_map_t::iterator iiso_iter = iso_iter->iso_map.begin(); 
	    iiso_iter != iso_iter->iso_map.end(); ++iiso_iter) {
	  VERB3 << iiso_iter->first << "->" << iiso_iter->second << " ";
	}
	VERB3 << "\nRoot: " << iso_iter->root << "\n---\n";
      }
    }
    mi = match_map.begin();
  }
  
  return std::make_pair(mi, mi_end);
}


bool Matcher::check(pvertex_id _u, pattern_t & _pattern, 
		    vertex_id _v, DAG & _dag, Node & _node,
		    TargetInstruction & _target_instruction,
		    std::vector< iso_map_s > & _iso_maps)
{
  node_map_t arg_map;

  //check if opcodes are equal
  if(!eq_ops_indeg(_u, _pattern,  _v, _dag, *(_node.get_node_attr()))) {
    return false;
  }
  //check residences of arguments
  if(!eq_arg_nodes(_u, _pattern, _v, _dag, _target_instruction, 
		   _node, arg_map)) {
    return false;
  }
  //check if out-going edges are equal
  bool is_root = false;
  if(!eq_out_edges(_u, _pattern, _v, _dag, &is_root) || is_root)
    {
      if(is_root) {
	//occurs only for one to one mapping
	if(_iso_maps.size() == 0) {
	  _iso_maps.push_back(iso_map_s());
	}
	for(int i = 0; i < _iso_maps.size(); i++) {
	  _iso_maps[i].root = _v;
	  _iso_maps[i].arg_map = arg_map;
	}
	return true;
      }
      return false;
    }

  bool b_partial_iso = false;
  out_pedge_iter pi, pi_end;
  pedge_type_map::type petm = edge_type_property(_pattern);  
  std::vector< pvertex_id > p_kid0_v, p_kid1_v, p_ddep_v, p_link_v;
  //Order pattern vertices according to their connecting edge type
  for(tie(pi, pi_end) = out_edges(_u, _pattern); pi != pi_end; ++pi) {
    pvertex_id p = target(*pi, _pattern);
    switch(petm[*pi]) {

    case KID0:
      p_kid0_v.push_back(p);
      break;

    case KID1:
      p_kid1_v.push_back(p);
      break;

    case DDEP:
      p_ddep_v.push_back(p);
      break;

    case LINK:
      p_link_v.push_back(p);
      break;

    default: 
      assert(false);
    }
  }

  std::vector< vertex_id > q_kid0_v, q_kid1_v, q_ddep_v, q_link_v;
  out_edge_iter qi, qi_end;
  edge_type_map::type qetm = edge_type_property(_dag);
  for(tie(qi, qi_end) = out_edges(_v, _dag); qi != qi_end; ++qi) {
    vertex_id q = target(*qi, _dag);
    switch(qetm[*qi]) {
    case KID0:
      q_kid0_v.push_back(q);
      break;

    case KID1:
      q_kid1_v.push_back(q);
      break;

    case DDEP:
      q_ddep_v.push_back(q);
      break;

    case LINK:
      q_link_v.push_back(q);
      break;

    default: 
      assert(false);
    }
  }

  std::sort(p_kid0_v.begin(), p_kid0_v.end());
  std::sort(p_kid1_v.begin(), p_kid1_v.end());
  std::sort(p_ddep_v.begin(), p_ddep_v.end());
  std::sort(p_link_v.begin(), p_link_v.end());

  std::vector< iso_map_s > partial_iso_maps;

  //For all permutations
  do {
    do {
      do {
	do {
	  for(int i = 0; i < q_kid0_v.size(); i++) {
	    if(!check(p_kid0_v[i], _pattern, q_kid0_v[i], _dag, _node,
		      _target_instruction, partial_iso_maps)) {
	      goto stop_current_check;
	    }
	  }
	  for(int i = 0; i < q_kid1_v.size(); i++) {
	    if(!check(p_kid1_v[i], _pattern, q_kid1_v[i], _dag, _node,
		      _target_instruction, partial_iso_maps)) {
	      goto stop_current_check;
	    }
	  }
	  for(int i = 0; i < q_ddep_v.size(); i++) {
	    if(!check(p_ddep_v[i], _pattern, q_ddep_v[i], _dag, _node, 
		      _target_instruction, partial_iso_maps)) {
	      goto stop_current_check;
	    }
	  }
	  for(int i = 0; i < q_link_v.size(); i++) {
	    if(!check(p_link_v[i], _pattern, q_link_v[i], _dag, _node,
		      _target_instruction, partial_iso_maps)) {
	      goto stop_current_check;
	    }
	  }
	  b_partial_iso = true;

	  for(std::vector< iso_map_s >::iterator pmi = partial_iso_maps.begin();
	      pmi != partial_iso_maps.end(); ++pmi) {
	    _iso_maps.push_back(iso_map_s(*pmi));
	    iso_map_s & part_iso = _iso_maps.back();
	    //extend with current matched nodes
	    for(int i = 0; i < q_kid0_v.size(); i++) {
	      part_iso.iso_map[ p_kid0_v[i] ] = q_kid0_v[i];
	    }
	    for(int i = 0; i < q_kid1_v.size(); i++) {
	      part_iso.iso_map[ p_kid1_v[i] ] = q_kid1_v[i];
	    }
	    for(int i = 0; i < q_ddep_v.size(); i++) {
	      part_iso.iso_map[ p_ddep_v[i] ] = q_ddep_v[i];
	    }
	    for(int i = 0; i < q_link_v.size(); i++) {
	      part_iso.iso_map[ p_link_v[i] ] = q_link_v[i];
	    }
	    //Arguments
	    for(node_map_t::iterator arg_iter = arg_map.begin(); 
		arg_iter != arg_map.end(); ++arg_iter) {
	      part_iso.arg_map[arg_iter->first] = arg_iter->second;
	    }
	  }

	stop_current_check:
	  //checked failed, go to next permutation
	  partial_iso_maps.clear();
	  
	} while(std::next_permutation(p_kid0_v.begin(), p_kid0_v.end()));
      } while(std::next_permutation(p_kid1_v.begin(), p_kid1_v.end()));
    } while(std::next_permutation(p_ddep_v.begin(), p_ddep_v.end()));
  } while(std::next_permutation(p_link_v.begin(), p_link_v.end()));

  return b_partial_iso;
}

bool Matcher::eq_ops_indeg(pvertex_id _u, pattern_t & _pattern, 
			   vertex_id _v, DAG & _dag, NodeAttr & _node_attr)
{
  vertex_info_map::type vim = vertex_info_property(_pattern);
  
  if(_node_attr.indeg[_v] != vim[_u].indeg) {
    return false;
  }
  std::list< int > & ops = vim[_u].ops;
  int v_op = (get_lcc_node(_v, _dag))->op;
  for(std::list< int >::iterator ops_iter = ops.begin(); 
      ops_iter != ops.end(); ++ops_iter) {
    if(*ops_iter == v_op) {
      return true;
    }
  }
  return false;
}

bool Matcher::eq_out_edges(pvertex_id _u, pattern_t & _pattern,
			   vertex_id _v, DAG & _dag, bool *_is_root)
{
  pedge_type_map::type petm = edge_type_property(_pattern);
  edge_type_map::type qetm = edge_type_property(_dag);
  int u_nb_edges[4] = { 0, 0, 0, 0 };
  int v_nb_edges[4] = { 0, 0, 0, 0 };
  out_pedge_iter pi, pi_end;
  out_edge_iter qi, qi_end;

  *_is_root = true;
  for(tie(pi, pi_end) = out_edges(_u, _pattern); pi != pi_end; ++pi) {
    u_nb_edges[ petm[*pi] ]++;
    *_is_root = false;
  }
  for(tie(qi, qi_end) = out_edges(_v, _dag); qi != qi_end; ++qi) {
    v_nb_edges[ qetm[*qi] ]++;
  }
  for(int i = 0; i < 4; i++) {
    if(u_nb_edges[i] != v_nb_edges[i]) {
      return false;
    }
  }
  return true;
}

bool Matcher::eq_arg_nodes(pvertex_id & _u, pattern_t & _pattern,
			   vertex_id & _v, DAG & _dag,
			   TargetInstruction & _target_instruction,
			   Node & _node,
			   node_map_t & _arg_map)
{

//#ifndef _SP_
  //return true;
//#endif //_SP_

  std::pair< in_pedge_iter, in_pedge_iter > p = in_edges(_u, _pattern);
  pedge_type_map::type petm = edge_type_property(_pattern);
  vertex_info_map::type pvim = vertex_info_property(_pattern);
  vertex_kind_map::type pvkm = vertex_kind_property(_pattern);
  edge_type_map::type etm = edge_type_property(_dag);
  std::pair< in_edge_iter, in_edge_iter > q = in_edges(_v, _dag);

  int parg_rc[2] = { -1, -1 };
  pvertex_id kids_u[2];
  vertex_id kids_v[2];
  bool b_arg = false;

  //Checking only KID0 and KID1 edges
  for( ; p.first != p.second; ++p.first) {
    pvertex_id kid_u = source(*(p.first), _pattern);
    if(pvkm[kid_u] == ARGNODE) {
      enum EDGE_T et = petm[*(p.first)];

      //debug
      //       VERB1 << "RC : " << pvim[kid_u].id << "\n";
      //       VERB1 << "TINST:\n" << (void *)&_target_instruction 
      // 		<< " " << _target_instruction << std::endl;
      //

      //debug
      //      std::cout << _target_instruction << std::endl;
      //debug

      parg_rc[et] = _target_instruction.get_residence_class(pvim[kid_u].id);
      kids_u[et] = kid_u;
      b_arg = true;
    }
  }

  //No argument nodes
  if(!b_arg) {
    return true;
  }

  for( ; q.first != q.second; ++q.first) {
    enum EDGE_T et;
    switch((et = etm[*(q.first)])) {  
    case KID0: case KID1:
      {
	if(parg_rc[et] != -1) {
	  vertex_id kid_v = source(*(q.first), _dag);
	  std::pair< bool, std::set< int > > rcs = _node.get_residence_class(kid_v);
	  if(!rcs.first) {
	    return false;
	  }
	  bool b_found = false;
	  //debug
	  // 	  VERB3 << "RC: ";
	  // 	  std::copy(rcs.second.begin(), rcs.second.end(), 
	  // 		    std::ostream_iterator< int >(VERB1, " "));
	  // 	  VERB3 << " - lookup: " << parg_rc[et] << "\n";
	  // 	  //
	  if(rcs.second.find(parg_rc[et]) != rcs.second.end()) {
	    b_found = true;
	    kids_v[et] = kid_v;
	    break;
	  }
	  if(!b_found) {
	    return false;
	  }
	}
      }
    }
  }
  //set mapping in iso_map for args
  for(int j = 0; j < 2; j++) {
    if(parg_rc[j] != -1) {
      if(_arg_map.find(kids_u[j]) != _arg_map.end()) {
	if(_arg_map[kids_u[j]] != kids_v[j]) {
	  return false;
	}
      }
      _arg_map[kids_u[j]] = kids_v[j];
    }
  }
  return true;
}
