#ifndef __MATCHER_HH__
#define __MATCHER_HH__

#include <map>
#include <vector>
#include "node.hh"
#include "archi.hh"
#include "pattern.hh"

class Matcher {

public:
  Matcher(Architecture *);
  Matcher(const Matcher & _m);
  ~Matcher();

  typedef std::map< TargetInstruction *, std::pair< pattern_t *, std::vector< iso_map_s > > > match_map_t;
  typedef match_map_t::iterator match_map_iter;

  std::pair< match_map_iter, match_map_iter >
  matches(vertex_id & , Node &, DAG &);

private:

  match_map_t match_map;
  Architecture *architecture;

  //pattern checking
  bool check(pvertex_id _u, pattern_t & _pattern,
	     vertex_id _v, DAG & _dag, Node & _node,
	     TargetInstruction & _target_instruction,
	     std::vector< iso_map_s > & _iso_maps);
  bool eq_ops_indeg(pvertex_id _u, pattern_t & _pattern, 
		    vertex_id _v, DAG & _dag, NodeAttr & _node_attr);
  bool eq_out_edges(pvertex_id _u, pattern_t & _pattern,
		    vertex_id _v, DAG & _dag, bool *_is_root);
  bool eq_arg_nodes(pvertex_id & _u, pattern_t & _pattern,
		    vertex_id & _v, DAG & _dag,
		    TargetInstruction & _target_instruction,
		    Node & _node,
		    node_map_t & _arg_map);


};

#endif //__MATCHER_HH__
