#ifndef __OPTS_HH__
#define __OPTS_HH__

#include "opt.hh"
#include "pudag.hh"
#include "zset.hh"
#include "sched.hh"
#include "grid.hh"
#include "matcher.hh"
#include "snode.hh"
#include "flags.hh"

class Optimist : public Optimizer {

public:

  Optimist(Architecture *);
  ~Optimist();
  Schedule *schedule(DAG &, Flags &);
	
private:

  static const int GRID_DIM = 2; 
  static const int T_AXIS = 0;
  static const int L_AXIS = 1;

  typedef grid< SNode *, GRID_DIM > solution_space_t;
  typedef std::vector< unsigned int > coord_t;
#ifdef __GNUC__
  typedef __gnu_cxx::hash_set< SNode *, __gnu_cxx::hash< SNode * >, std::equal_to< SNode * > > grid_elem_t;
#else
  typedef std::hash_set< SNode *, std::hash< SNode * >, std::equal_to< SNode * > > grid_elem_t;
#endif	
  NodeAttr *get_init_node_attr(DAG &);
  int reduce_init_snode(SNode *, DAG &);
  NodeAttr *select(const vertex_set_t & chi, const NodeAttr &, DAG &);
 
  void deallocate_nodes(solution_space_t & solution_space,
			int num_vertices, unsigned int t);
  void match_node(vertex_id u_id, SNode * current_snode, 
			    DAG & _dag, coord_t & coord, 
			    solution_space_t & solution_space);
  void transfers(Flags & _flags, SNode * current_snode,
                 DAG & _dag, coord_t & coord,
                 solution_space_t & solution_space);

  void update_space(const SNode *, TargetInstruction *,
		    const vertex_set_t &, const coord_t &, 
		    solution_space_t & _solution_space,
		    iso_map_s & _iso_maps, 
		    pattern_t & _pattern, DAG _dag);

  coord_t max_coord;
  ZSet *zset;

  //Statistics
  int cut, improved, merged, n_snodes;

  Architecture *architecture;
  Matcher *matcher;
  coord_t last_in;

  typedef bool (*condition)(std::vector< lcc::Node >);
  condition *conditions;

  /* Classification statistics */
  std::vector< unsigned int > nr_greedy;
  std::vector< unsigned int > nr_str_lin;

  void classify_solutions(coord_t coord, DAG & _dag, solution_space_t & _solution_space);
};

#endif //__OPTS_HH__
