#include "sched.hh"
#include "node.hh"

#define ping std::cout << "*ping*" << __LINE__ << "\n";
#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b)) 

/* helper function */
void print_iso_map(iso_map_s & im)
{
  std::cout << "iso_map_s contents: \n";
  std::cout << "sizes: im " << im.iso_map.size()
	    << ", am " << im.arg_map.size()
	    << ", root is " << im.root << std::endl;

}

Schedule::Schedule(int _omega)
  : omega(_omega)
{
  s.push_back(std::vector< slot_t >(_omega));
  std::vector< slot_t > & tmp = s.back();
  for(int i = 0; i < omega; i++) {
    tmp[i].first = NULL;
  }
  //roots.push_back(std::vector< vertex_id >(_omega));
  im.push_back(std::vector< iso_map_s >(_omega));
  for(int i=0; i<_omega; ++i)
    (im.back())[i].root = 999999999;
}

std::pair< bool, int > 
Schedule::insert(TargetInstruction * _target_inst,
		 std::map< char *, std::pair< char *, lcc::Node>, ltstr > & _inst_instance,
		 pattern_t *_pattern, iso_map_s _iso_maps, const vertex_set_t &_chi, int _offset)
{
//   vertex_set_t::iterator j = _chi.end();
//   for(vertex_set_t::iterator i = _chi.begin(); i!= j; ++i)
//     std::cout << "CHI: " << _chi.size() << " ";
//   std::cout << std::endl;
  advance(_offset);
  for(int i = 0; i < omega; i++) {
    std::vector< slot_t > & tmp = s.back();
    //std::vector< vertex_id > & tmp2 = roots.back();
    std::vector< iso_map_s > & tmp3 = im.back();
    if(tmp[i].first == NULL) {
      //empty slot
      //tmp2[i] = _iso_maps.root;
      tmp3[i] = _iso_maps;

      tmp[i].first = _pattern;
      tmp[i].second.first = _target_inst;
      tmp[i].second.second = _inst_instance;
      return std::make_pair(true, 0);
    }
  }

  s.push_back(std::vector< slot_t >(omega));
  //roots.push_back(std::vector< vertex_id >(omega));
  im.push_back(std::vector< iso_map_s >(omega));
  for(int i=0; i<omega; ++i)
    (im.back())[i].root = 999999999;

  std::vector< slot_t > & tmp = s.back();
  //std::vector< vertex_id > & tmp2 = roots.back();
  std::vector< iso_map_s > & tmp3 = im.back();
  for(int j = 0; j < omega; j++) {
    tmp[j].first = NULL;
  }
  //tmp2[0] = _iso_maps.root;
  tmp3[0] = _iso_maps;

  tmp[0].first = _pattern;
  tmp[0].second.first = _target_inst;
  tmp[0].second.second = _inst_instance;

  return std::make_pair(true, 1);
}

Schedule::~Schedule()
{
}

Schedule::Schedule(const Schedule & _s)
{
  omega = _s.omega;
  s = _s.s;
  //roots = _s.roots;
  im = _s.im;
  //iso_maps = _s.iso_maps;
}

std::ostream & operator<<(std::ostream & _out, const Schedule & _s)
{
  int length = _s.s.size();
  int omega = _s.s[0].size();
  _out << "Schedule size: " << length << "\n";
  std::vector< std::vector< Schedule::slot_t > > s = (_s.s);
  for(int i = 0; i < length; i++) {
    for(int j = 0; j < omega; j++) {
      Schedule::slot_t & slot = s[i][j];
      if(slot.first) {
	for(std::map< char *, std::pair< char *, lcc::Node >, ltstr >::const_iterator iter = slot.second.second.begin();
	    iter != slot.second.second.end(); ++iter) {
	  char *str1 = iter->first;
	  char *str2 = iter->second.first;
	  _out << _s.im[i][j].root << " ";
	  _out << "[" << i << "," << j << "]" << str1 << ":" << str2 << std::endl;
	}
      }
    }
  }
  return _out;
}

void Schedule::advance(int _offset)
{
  if(_offset > 0) {
    for(int i = 0; i < _offset; i++) {
      s.push_back(std::vector< slot_t >(omega));
      //roots.push_back(std::vector< vertex_id >(omega));
      im.push_back(std::vector< iso_map_s >(omega));
      for(int i=0; i<omega; ++i)
	(im.back())[i].root = 999999999;

      std::vector< slot_t > & tmp = s.back();
      for(int j = 0; j < omega; j++) {
	tmp[j].first = NULL;
      }
    }
  }
}

bool Schedule::available_slot(void)
{
  std::vector< slot_t > & tmp = s.back();
  for(int i = 0; i < omega; i++) {
    if(tmp[i].first == NULL) {
      return true;
    }
  }
  return false;
}

/* mater impl. */
void Schedule::print_reserv_tab(const reserv_tab_t & t) 
{
  for(reserv_tab_t::const_iterator k = t.begin(); /* resource */
      k != t.end();
      ++k){
    for(bit_vector_t::const_iterator l = k->begin(); /* time */
	l != k->end();
	++l){
      std::cout << (*l?"B":"_");
    }
    std::cout << std::endl;
  }
}

void Schedule::print_resource_usage(Architecture* architecture)
{
  int time = 0;
  reserv_tab_t full_res_tab;
  for(int i=0; i<architecture->get_reserv_tab_size(); i++){
    bit_vector_t tmp;
    full_res_tab.push_back(tmp);
  }
  std::cout << "The resource usage: \n";
  /* for each time slot */
  for(std::vector< std::vector< slot_t > >::iterator i = s.begin();
      i != s.end();
      ++i,++time){
    /* for each instruction issued at the time slot */
    for(std::vector<slot_t>::iterator j = i->begin();
	j!=i->end();
	++j){
      TargetInstruction* tinst = (j->second.first);
      if(tinst != NULL){
	reserv_tab_t new_usage = tinst->get_reserv_tab();
	merge_reserv_tab(full_res_tab, new_usage, time);
      }
    }
  }
  print_reserv_tab(full_res_tab);
}

/* new_usage and offset should be const */
void Schedule::merge_reserv_tab(reserv_tab_t & base, const reserv_tab_t & new_usage, int offset)
{
  reserv_tab_t::iterator base_resource_it = base.begin();
  reserv_tab_t::const_iterator new_resource_it;
  bit_vector_t::iterator base_time_it;
  bit_vector_t::const_iterator new_time_it;

  assert( base_resource_it != base.end());
    
  /* For each resource add new usage */
  for(new_resource_it = new_usage.begin();
      new_resource_it != new_usage.end();
      ++new_resource_it, ++base_resource_it){
    /* append time slots until we have offset+size of new usage*/
    while(base_resource_it->size() < new_resource_it->size() + offset)
      base_resource_it->push_back(false);

    /* loop over time */
    for(base_time_it = base_resource_it->begin()+offset, new_time_it = new_resource_it->begin();
	new_time_it != new_resource_it->end();
	++base_time_it, ++new_time_it){
      assert(!(*base_time_it && *new_time_it));
      *base_time_it = (*base_time_it || *new_time_it);
    }
  }
}

bool Schedule::insertable(reserv_tab_t & base, const reserv_tab_t & new_usage, int offset)
{
  reserv_tab_t::iterator base_resource_it = base.begin();
  reserv_tab_t::const_iterator new_resource_it;
  bit_vector_t::iterator base_time_it;
  bit_vector_t::const_iterator new_time_it;

  assert( base_resource_it != base.end());
    
  /* For each resource add new usage */
  for(new_resource_it = new_usage.begin();
      new_resource_it != new_usage.end();
      ++new_resource_it, ++base_resource_it){
    /* append time slots until we have offset+size of new usage*/
    while(base_resource_it->size() < new_resource_it->size() + offset)
      base_resource_it->push_back(false);

    /* loop over time */
    for(base_time_it = base_resource_it->begin()+offset, new_time_it = new_resource_it->begin();
	new_time_it != new_resource_it->end();
	++base_time_it, ++new_time_it){
      if (*base_time_it && *new_time_it)
	return false;
    }
  }
  return true;
}

void Schedule::print_stuff(DAG & _dag)
{
  int time = 0;
  std::cout << "Whole dag size " << num_vertices(_dag) << std::endl;
  std::cout << "pattern dag size\n";
  /* for each time slot */
  for(std::vector< std::vector< slot_t > >::iterator i = s.begin();
      i != s.end();
      ++i,++time){
    /* for each instruction issued at the time slot */
    for(std::vector<slot_t>::iterator j = i->begin();
	j!=i->end();
	++j){
      if (j->first != NULL)
	std::cout << "time " << time << " " << num_vertices(*(j->first)) << std::endl;
    }
  }
}

/* only works when each covering is 1 operator and a few arguments */
/* eg. multiply-and-accumulate does not work! */
bool Schedule::greedy(DAG & _dag, Architecture *architecture)
{
  int time = 0, unit = 0;
  /* new empty resource usage map */
  reserv_tab_t full_res_tab;
  for(int i=0; i<architecture->get_reserv_tab_size(); i++){
    bit_vector_t tmp;
    full_res_tab.push_back(tmp);
  }
  /* New empty latency map (This name is misleading and should be
     changed!) */
  std::map< vertex_id, int > latency_map;

  /* for each time slot */
  for(std::vector< std::vector< slot_t > >::iterator i = s.begin();
      i != s.end();
      ++i,++time){
    /* for each instruction issued at the time slot */
    unit = 0;
    for(std::vector<slot_t>::iterator j = i->begin();
	j!=i->end(), j->first != NULL;
	++j, ++unit){
      slot_t & current_slot = *j;
      pattern_t & new_pattern = *(j->first);
      /* calculate offset wrt latencies of predecessors */
      int offset = 0;
      /* create the set of new nodes */ 
      vertex_set_t new_nodes;
      iso_map_s & new_im = im[time][unit];
      vertex_id new_root = new_im.root;
      new_nodes.insert(new_root);
      node_map_t::iterator args = new_im.arg_map.begin();
      node_map_t::iterator args_end = new_im.arg_map.end();
      for(;args != args_end; ++args){
	//new_nodes.insert((*args).second);
      }

      /* for each vertex in new_nodes */
      vertex_set_t::iterator nit1, nit2;
      for(nit1=new_nodes.begin(),
	    nit2=new_nodes.end();
	  nit1!=nit2;
	  ++nit1){
	/* for each vertex in predecessors of that vertex */
	predecessor_iterator pre, pre_end;
	for(tie(pre, pre_end) = predecessors(*nit1, _dag);
	    pre != pre_end;
	    ++pre){
	  vertex_id v_in_dag = *pre;
	  offset = max(offset, latency_map[v_in_dag]);
	  if(offset>time){
	    /* this should never happen */
	    std::cout << "XXX - ERROR, schedule impossible?" 
		      << " Offset is " << offset << " and scheduled at " << time
		      << "\n pre:" << v_in_dag
		      << std::endl;
	    std::cout << *this << std::endl;
	    print_reserv_tab(full_res_tab);
	    //	  exit(1);
	  }
	}
      }

      /* increase offset until the reservation table fits */
      TargetInstruction* tinst = (j->second.first);
      if(tinst != NULL){
	const reserv_tab_t & new_usage = tinst->get_reserv_tab();
	while(!(insertable(full_res_tab, new_usage, offset)))
	  offset++;
	if(offset>time){
	  /* this should never happen */
	  std::cout << "ERROR, schedule impossible?" 
		    << " Offset is" << offset << "and scheduled at " << time
		    << std::endl;
	  std::cout << *this << std::endl;
	  print_reserv_tab(full_res_tab);
	  print_reserv_tab(new_usage);
	  exit(1);
	}
	if(offset<time){ /* We found a non-greedy schedule */
// 	  tinst->print();
// 	  std::cout << " time: " << time << ",root: " << new_root << std::endl;

	  return false;
	}
	/* Now commit to full_res_tab and latency_map */
	merge_reserv_tab(full_res_tab, new_usage, time);
	/* get latency of new instruction */
	int latency = 0;
	//tinst->print();
	//print_reserv_tab(new_usage);
	for(reserv_tab_t::const_iterator rit = new_usage.begin();
	    rit != new_usage.end();
	    ++rit){
	  latency = max(latency, rit->size());
	}
// 	std::cout << "Scheduling " << new_root 
// 		  << "(";
// 	tinst->print();
// 	std::cout << ")"
// 		  << " @ " << time << std::endl;
// 	std::cout << "Latency " << new_root << " : " << latency+time << std::endl;
	latency_map[new_root] = latency+time;
      }
    }
  }
  /* We did not find a non-greedy instruction */
  return true;
}

bool Schedule::operator!=(const Schedule & _sched){
  return (!(*this == _sched));
}
 
bool Schedule::operator==(const Schedule & _sched)
{
  /* Check if the two schedules are equivalent by checking issuing (nodes) */
  typedef std::vector< std::vector < iso_map_s > > imvv;
  typedef std::vector< iso_map_s > imv;

  /* if not same size they can not be equivalent */
//   if(im.size() != _sched.im.size())
//     return false;

  imvv::const_iterator a_it, a_end, b_it, b_end;
  a_it = im.begin();
  b_it = _sched.im.begin();
  a_end = im.end();
  b_end = _sched.im.end();

  imv::const_iterator aa_it, aa_end, bb_it, bb_end;

  for(; (a_it != a_end) && (b_it != b_end); 
      ++a_it, ++b_it){
//     ((*a_it).size() != (*b_it).size())
//       return false;
    aa_it = a_it->begin();
    bb_it = b_it->begin();
    aa_end = a_it->end();
    bb_end = b_it->end();
    for(; (aa_it != aa_end) && (bb_it != bb_end); 
	++aa_it, ++bb_it){
      /* Now we are at iso_map level, check if the roots are equal */
      if(aa_it->root != bb_it->root){
	return false;
      }
    }
  }
  return true;
}

/* This is not working because iso_maps is not complete */
/* it does not hold the full set of vertices */
// bool Schedule::greedy(DAG & _dag, Architecture *architecture)
// {
//   int time = 0;
//   /* new empty resource usage map */
//   reserv_tab_t full_res_tab;
//   for(int i=0; i<architecture->get_reserv_tab_size(); i++){
//     bit_vector_t tmp;
//     full_res_tab.push_back(tmp);
//   }
//   /* new empty latency map */
//   std::map< vertex_id, int > latency_map;

//   /* for each time slot */
//   for(std::vector< std::vector< slot_t > >::iterator i = s.begin();
//       i != s.end();
//       ++i,++time){
//     /* for each instruction issued at the time slot */
//     for(std::vector<slot_t>::iterator j = i->begin();
// 	j!=i->end(), j->first != NULL;
// 	++j){
//       slot_t & current_slot = *j;
//       pattern_t & new_pattern = *(j->first);
//       /* calculate offset wrt latencies of predecessors */
//       int offset = 0;
//       predecessor_iterator pre, pre_end;
//       pvertex_iter new_vertex, new_vertex_end;
//       /* create the set of new nodes */
//       vertex_set_t new_nodes;
//       for(tie(new_vertex, new_vertex_end) = vertices(new_pattern);
// 	  new_vertex != new_vertex_end;
// 	  ++new_vertex){
// 	new_nodes.insert( (iso_maps[&current_slot].iso_map)[*new_vertex] );
// 	node_map_t::iterator tmpa, tmpb;
// 	for(tmpa=(iso_maps[&current_slot].iso_map).begin(),
// 	      tmpb=(iso_maps[&current_slot].iso_map).end();
// 	    tmpa != tmpb;
// 	    ++tmpa){
// 	  std::cout << tmpa->first << " maps to " << tmpa->second << "\n";
// 	}
// 	std::cout << "Inserted vertex ID (dag id) " 
// 		  <<  (iso_maps[&current_slot].iso_map)[*new_vertex] << "\n";
//       }
//       std::cout << "contents of latency_map is :\n";
//       std::map< vertex_id, int >::iterator tmp1, tmp2;
//       for(tmp1 = latency_map.begin(), tmp2 = latency_map.end(); tmp1!=tmp2;++tmp1){
// 	std::cout << tmp1->first << ":" << tmp1->second << " ";
//       }
//       std::cout << std::endl;
//       /* for each vertex in the new nodes */
//       for(tie(new_vertex, new_vertex_end) = vertices(new_pattern);
// 	  new_vertex != new_vertex_end;
// 	  ++new_vertex){
// 	vertex_id u = *new_vertex;
// 	vertex_id u_in_dag = (iso_maps[&current_slot].iso_map)[u];
// 	/* for each vertex in predecessors */
// 	for(tie(pre, pre_end) = predecessors(u_in_dag, _dag);
// 	    pre != pre_end;
// 	    ++pre){
// 	  vertex_id v_in_dag = *pre;
	  
// 	  /* if the predecessor is not in the set of new nodes increase offset */
// 	  if(new_nodes.find(v_in_dag) == new_nodes.end()){
// 	    std::cout << "Found a predecessor with latency " << latency_map[v_in_dag] << std::endl;
// 	    offset = max(offset, latency_map[v_in_dag]);
// 	  }
// 	}
//       }
//       std::cout << "offset wrt latencies " << offset << std::endl;


//       /* increase offset until the reservation table fits */
//       TargetInstruction* tinst = (j->second.first);
//       if(tinst != NULL){
// 	reserv_tab_t new_usage = tinst->get_reserv_tab();
// 	while(!(insertable(full_res_tab, new_usage, offset)))
// 	  offset++;
// 	std::cout << " earliest possible time " << offset 
// 		  << " scheduled at " << time << std::endl << std::endl;
// 	/* Now commit to full_res_tab and latency_map */
// 	merge_reserv_tab(full_res_tab, new_usage, time);
// 	/* get latency of new instruction */
// 	int latency = 0;
// 	tinst->print();
// 	print_reserv_tab(new_usage);
// 	for(reserv_tab_t::iterator rit = new_usage.begin();
// 	    rit != new_usage.end();
// 	    ++rit){
// 	  latency = max(latency, rit->size());
// 	}
// 	vertex_set_t::iterator new_nodes_in_dag_it, new_nodes_in_dag_end;
// 	for(new_nodes_in_dag_it = new_nodes.begin(), new_nodes_in_dag_end = new_nodes.end();
// 	    new_nodes_in_dag_it != new_nodes_in_dag_end;
// 	    ++new_nodes_in_dag_it){
// 	  std::cout << "INSERTING LATENCY " << latency+time << " for node " 
// 		    << *new_nodes_in_dag_it << "\n\n";
	  
// 	  //if(latency_map.find(*new_nodes_in_dag_it) != latency_map.end())
// 	  //  std::cout << "ERROR, node already in latency map " << *new_nodes_in_dag_it << "\n";
// 	  latency_map[*new_nodes_in_dag_it] = latency+time;
// 	}
//       }
//     }
//   }
// }
