#include <assert.h>
#include "snode.hh"
#include <algorithm>

#define max(a,b) (((a)>(b))?(a):(b))

SNode::SNode(NodeAttr * _node_attr, TimeProfile *_time_profile,
	     Architecture *_architecture)
  : node_attr(_node_attr), reference_time(0)
{
  merged = 1;
  is_greedy = true;
  //empty schedule
  time_profile = new TimeProfile(*_time_profile);
  s = new Schedule(_architecture->get_omega());
#ifdef _SP_
  space_profile = new SpaceProfile(*_architecture);
#else
  available_map.clear();
#endif //_SP_
}

SNode::SNode(const SNode & _snode)
{
  *this = _snode;
}

SNode::~SNode()
{
  delete s;
  s = NULL;
  delete time_profile;
  time_profile = NULL;
#ifdef _SP_
  delete space_profile;
  space_profile = NULL;
#else
  available_map.clear();
#endif //_SP_
}

SNode & SNode::operator=(const SNode & _snode)
{
  merged = _snode.get_merged();
  is_greedy = _snode.greedy();
  node_attr = _snode.node_attr;
  reference_time = _snode.reference_time;
  time_profile = new TimeProfile(*(_snode.time_profile));
  s = new Schedule(*(_snode.s));
#ifdef _SP_
  space_profile = new SpaceProfile(*(_snode.space_profile));
#else
  available_map = _snode.available_map;
#endif //_SP_
  return *this;
}

bool SNode::operator==(const SNode & _snode) const
{
#if 0 
  /* This was used when we did the cpc paper on classification. */
  /* Idea: we can merge two snodes if they both are non-greedy.  The
  cost of this is calculating for every partial schedule if it is
  greedy or not. */

  if(is_greedy && _snode.greedy()){
    return (*s == *_snode.get_schedule());
  }
  if(is_greedy || _snode.greedy())
    return false;
  if(reference_time != _snode.get_reference_time())
    return false;
  /* The above is only for classification. */
#endif

  return (_snode.node_attr == node_attr) && 
    ((*(_snode.time_profile)) == (*time_profile)) 
#ifdef _SP_
    &&
    (*_snode.space_profile) == (*space_profile)
#endif //_SP_
    ;
}

//Return new reference time of the time profile
std::pair< bool, int >
SNode::insert(const vertex_set_t & _chi,
	      //vertex_id & _root, 
	      DAG & _dag, 
	      iso_map_s & _iso_maps,
	      //	      node_map_t & _arg_map, 
	      pattern_t & _pattern,
	      TargetInstruction *_target_instruction,
	      Architecture * _architecture)
{
 
  int additional_offset = 0;
  bool b_inserted;
  int root_rc, destid_rc = -1;
  std::map< char *, std::pair< char *, lcc::Node >, ltstr > _ops_mapping;
  vertex_info_map::type vim = vertex_info_property(_pattern);

  node_map_t _arg_map = _iso_maps.arg_map;

#ifdef _SP_
  //Arguments


  for(node_map_t::iterator arg_iter = _arg_map.begin(); 
      arg_iter != _arg_map.end(); ++arg_iter) {

    char *arg_id = vim[arg_iter->first].id;
    int residence_class = _target_instruction->get_residence_class(arg_id);
    int earliest_instance = -1;
    std::pair< bool, std::pair< char *, lcc::Node > > ip = 
      space_profile->get_instance(arg_iter->second, residence_class,
				  reference_time, _architecture,
				  &earliest_instance);
    if(!ip.first) {
      if(earliest_instance == -1) {
	return std::make_pair(false, 0);
      }
      //Need insert NOPs.
      int NOP_offset = (earliest_instance - reference_time);
      assert(NOP_offset > 0);
      //Time profile
      time_profile->advance(NOP_offset);
      //Schedule
      s->advance(NOP_offset);
      reference_time += NOP_offset;
    }
    //Map arguments' residence class with an instance
    _ops_mapping[arg_id] = ip.second;
  }
#else
  int earliest_instance = -1;
  for(node_map_t::iterator arg_iter = _arg_map.begin(); 
      arg_iter != _arg_map.end(); ++arg_iter) {

    if(available_map.find(arg_iter->second) == available_map.end()) {
      return std::make_pair(false, 0);
    }
    earliest_instance = 
      max(earliest_instance, available_map[arg_iter->second]);
    
    int NOP_offset = (earliest_instance - reference_time);
    assert(NOP_offset > 0);
    //Time profile
    time_profile->advance(NOP_offset);
    //Schedule
    s->advance(NOP_offset);
    reference_time += NOP_offset;
  }
#endif //_SP_

  const reserv_tab_t & reserv_tab = _target_instruction->get_reserv_tab();
  //If there is no available slot and if we need to issue a target instruction, advance by one in schedule.
  if((!s->available_slot()) && ( strlen(_target_instruction->format) != 0)) {
    s->advance(1);
    time_profile->advance(1);
    reference_time += 1;
  }
  assert(s->available_slot() || strlen(_target_instruction->format) == 0);

  int tp_offset = time_profile->insert(_chi, reserv_tab, _dag);
  if(tp_offset > 0) {
    s->advance(tp_offset);
    reference_time += tp_offset;
  }

#ifdef _SP_
  //SP
  //This should be a field of reservation table. FIXME resrev_tab.height;
  int height = 0;
  for(int i = 0; i < reserv_tab.size(); i++) {
    height = max(reserv_tab[i].size(), height); 
  }

  std::pair< bool, std::pair< char *, lcc::Node > > p;

  //z.alive = alive U X (nodes under computation, i.e. tp.nodes)

 // FIXME
//   vertex_set_t alive;
//   std::insert_iterator< vertex_set_t >
//     alive_ins(alive, alive.begin());

//   std::set_difference(node_attr->alive.begin(), node_attr->alive.end(),
// 		      time_profile->get_nodes().begin(), time_profile->get_nodes().end(), 
// 		      alive_ins);
  //
//   std::cout << "zset.alive: ";
//   std::copy(node_attr->alive.begin(), node_attr->alive.end(), 
// 	    std::ostream_iterator< vertex_id >(std::cout, " "));
//   std::cout << "\nX: ";
//   std::copy(time_profile->get_nodes().begin(), time_profile->get_nodes().end(), 
// 	    std::ostream_iterator< vertex_id >(std::cout, " "));
//   std::cout << "\nalive: ";
//   std::copy(alive.begin(), alive.end(), 
// 	    std::ostream_iterator< vertex_id >(std::cout, " "));
//   std::cout << std::endl;
  //

  //
  //Implementation of constraint of resulting node.
  //root: if dest1_is specified, then the instance of the root resource is the same as of
  //node specified by dest1_is
  vertex_id dest_idx = (vertex_id)-1; //FIXME
  char *destid = _target_instruction->get_dest1_is();
  if( destid != 0) {
    vertex_info_map::type vim = vertex_info_property(_pattern);
    for(node_map_t::iterator arg_iter = _arg_map.begin();
	arg_iter != _arg_map.end(); ++arg_iter) {
      pvertex_id pvid = arg_iter->first;
      if(!strcmp(vim[pvid].id, destid)) {
	//Index of node 
	dest_idx = _arg_map[pvid];
	break;
      }
    }
    assert(dest_idx != (vertex_id)-1);
    if(_target_instruction->get_residence_class(destid) !=
       _target_instruction->get_residence_class("dest1")) {
      std::cerr << *(_target_instruction) << std::endl
		<< "Constraints does not hold:\nresidence class of the destination is different of the source.\n\n"
		<< std::endl;
      assert(false);
    }
    destid_rc = _target_instruction->get_residence_class(destid);
  }

  vertex_id root = _iso_maps.root;

  p = space_profile->insert(node_attr->alive, //FIXME alive, //node_attr->alive,
			    root, _dag,
			    (reference_time + height),
			    (root_rc = _target_instruction->get_residence_class("dest1")),
			    _architecture,
			    std::make_pair(dest_idx, destid_rc));
  if(!p.first) {
    return std::make_pair(false, 0);
  }

  //Destination of the root + mapping
  if(root_rc != STMT_RC) {
    _ops_mapping["dest1"] = p.second;
    if(num_vertices(_pattern) == 0 && _arg_map.size() == 0) {
      //transfer (src1 is the only argument - no pattern)
      int t_residence_class = _target_instruction->get_residence_class("src1");
      int earliest_instance = -1;
      std::pair< bool, std::pair< char *, lcc::Node > > ip = 
	space_profile->get_instance(root, t_residence_class, 
				    reference_time, _architecture, 
				    &earliest_instance);
      if(!ip.first) {
	return std::make_pair(false, 0);
      }
      _ops_mapping["src1"] = ip.second;
    }
  }
#else
  //This should be a field of reservation table. FIXME resrev_tab.height;
  int height = 0;
  for(int i = 0; i < reserv_tab.size(); i++) {
    height = max(reserv_tab[i].size(), height); 
  }
  vertex_id root = _iso_maps.root;
  available_map[root] = reference_time + height;
#endif //_SP_

  //Mapping
  node_map_t::iterator bi = _iso_maps.iso_map.begin(), 
    bi_end = _iso_maps.iso_map.end();

  for(; bi != bi_end; ++bi) {
    pvertex_id pvid = bi->first; //pvertex
    char *arg_id = vim[pvid].id;

    //Unamed vertexes of pattern are not considered
    if(std::string(arg_id) != "") {
      if(_ops_mapping.find(arg_id) == _ops_mapping.end()) {
	//add mapping
	vertex_id vid = bi->second;
	_ops_mapping[arg_id] = std::make_pair(arg_id, get_lcc_node(vid, _dag));
      }
    }
  }
  
  bi = _iso_maps.arg_map.begin(), 
    bi_end = _iso_maps.arg_map.end();

  for(; bi != bi_end; ++bi) {
    pvertex_id pvid = bi->first; //pvertex
    char *arg_id = vim[pvid].id;

    //Unamed vertexes of pattern are not considered
    if(std::string(arg_id) != "") {
      if(_ops_mapping.find(arg_id) == _ops_mapping.end()) {
	//add mapping
	vertex_id vid = bi->second;
	_ops_mapping[arg_id] = std::make_pair(arg_id, get_lcc_node(vid, _dag));
      }
    }
  }
  //Schedule: all mapping is done now

  //Advancing one slot in a time profile does not change its alive set
  if(strlen(_target_instruction->format) != 0) {
    tie(b_inserted, additional_offset) = 
      s->insert(_target_instruction, _ops_mapping, &_pattern, _iso_maps, _chi);
    assert(additional_offset == 0);
  }


  assert(s->size() == reference_time + 1);
#if 0
  /* schedule updated, check if it is greedy. It can only go from
     greedy to non-greedy, not the other way around. */
  if(is_greedy)
    is_greedy = s->greedy(_dag, _architecture);
#endif
  return std::make_pair(true, reference_time);
}

Schedule *SNode::get_schedule(void) const
{
  return s;
}

NodeAttr *SNode::get_node_attr(void) const
{
  return node_attr;
}

int SNode::get_reference_time(void) const
{
  return reference_time;
}

int SNode::get_tau(void) const
{
  return reference_time + time_profile->get_depth();
}

void SNode::set_reference_time(int _reference_time)
{
  reference_time = _reference_time;
}

void SNode::set_node_attr(NodeAttr *_node_attr)
{
  node_attr = _node_attr;
}

//Debug
std::ostream & operator<<(std::ostream & _out, const SNode & _snode)
{
  _out << *(_snode.node_attr);
  _out << *(_snode.time_profile);
#ifdef _SP_
  _out << *(_snode.space_profile);
#endif //_SP_
  _out << *(_snode.s);
  _out << "t: " << _snode.reference_time << " tau: " 
       << _snode.get_tau() << std::endl;
  return _out;
}
