#include <iterator>
#include <algorithm>
#include <assert.h>
#include "tp.hh"
#include "sched.hh" //for print_reserv_tab
#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b)) 

const size_t TimeProfile::primes[MP][MP] = {
     3,    5,    7,   11,   13,   17,   19,   23,   29,   31,   37,   41,   43,   47,   53, 
    59,   61,   67,   71,   73,   79,   83,   89,   97,  101,  103,  107,  109,  113,  127, 
   131,  137,  139,  149,  151,  157,  163,  167,  173,  179,  181,  191,  193,  197,  199, 
   211,  223,  227,  229,  233,  239,  241,  251,  257,  263,  269,  271,  277,  281,  283, 
   293,  307,  311,  313,  317,  331,  337,  347,  349,  353,  359,  367,  373,  379,  383, 
   389,  397,  401,  409,  419,  421,  431,  433,  439,  443,  449,  457,  461,  463,  467, 
   479,  487,  491,  499,  503,  509,  521,  523,  541,  547,  557,  563,  569,  571,  577, 
   587,  593,  599,  601,  607,  613,  617,  619,  631,  641,  643,  647,  653,  659,  661, 
   673,  677,  683,  691,  701,  709,  719,  727,  733,  739,  743,  751,  757,  761,  769, 
   773,  787,  797,  809,  811,  821,  823,  827,  829,  839,  853,  857,  859,  863,  877, 
   881,  883,  887,  907,  911,  919,  929,  937,  941,  947,  953,  967,  971,  977,  983, 
   991,  997, 1009, 1013, 1019, 1021, 1031, 1033, 1039, 1049, 1051, 1061, 1063, 1069, 1087, 
  1091, 1093, 1097, 1103, 1109, 1117, 1123, 1129, 1151, 1153, 1163, 1171, 1181, 1187, 1193, 
  1201, 1213, 1217, 1223, 1229, 1231, 1237, 1249, 1259, 1277, 1279, 1283, 1289, 1291, 1297, 
  1301, 1303, 1307, 1319, 1321, 1327, 1361, 1367, 1373, 1381, 1399, 1409, 1423, 1427, 1429, 
};

TimeProfile::TimeProfile(int _size) 
  : hash_value(0), current_time(0)
{
  rtab = new reserv_tab_t(_size);
  nodes.clear();
  ready_map.clear();
  latency_map.clear();
}

TimeProfile::~TimeProfile()
{
  delete rtab;
}

TimeProfile::TimeProfile(const TimeProfile & _tp)
  : 
    nodes(_tp.nodes), latency_map(_tp.latency_map),
    ready_map(_tp.ready_map),
    current_time(_tp.current_time)
{
  rtab = new reserv_tab_t(_tp.size());
  for(int i = 0; i < rtab->size(); i++) {
    (*rtab)[i] = _tp.get_rtab()[i];
  }
  hash_value = _tp.hash_value;
}

bool TimeProfile::operator==(const TimeProfile & _time_profile) const
{
  if(hash_value != _time_profile.hash_value) {
    return false;
  }
  if(nodes != _time_profile.nodes) {
    return false;
  }
  if(latency_map != _time_profile.latency_map) {
    return false;
  }
  int size = this->size();
  for(int i = 0; i < size; i++) {
    int v1_length = (*rtab)[i].size(), v2_length = _time_profile.get_rtab()[i].size();
    if(v1_length != v2_length) {
      return false;
    }
    for(int j = 0; j < v1_length; j++) {
      if((*rtab)[i][j] != _time_profile.get_rtab()[i][j]) {
	return false;
      }
    }
  }
  return true;
}

/******************************************************************************
 *                     TimeProfile::insert()                                  *
 * input:                                                                     *
 *  _nodes - set of nodes to be inserted                                      *
 *  _reserv_tab - resource usage                                              *
 *  _dag - the dag we are scheduling, used to find predecessors               *
 * note: when entering this function no kids of _nodes are in latency_map     *
 * (they are taken care of in snode.cc).                                      *
 ******************************************************************************/
int TimeProfile::insert(const vertex_set_t & _nodes,
			const reserv_tab_t & _reserv_tab,
			DAG & _dag)
{
  assert(_reserv_tab.size() == (*rtab).size());
  /* mater: advance until after latencies (data dependencies) */
  predecessor_iterator vi, vi_end;
  int offset = 0;
  int predecessors_ready = current_time;
  for(vertex_set_t::const_iterator ui = _nodes.begin(); ui != _nodes.end(); ++ui) {
    vertex_id u_id = *ui;
    for(tie(vi, vi_end) = predecessors(u_id, _dag); vi != vi_end; ++vi) {
      vertex_id v_id = *vi;
      if(ready_map.find(*vi) != ready_map.end()){
	predecessors_ready = max(predecessors_ready,ready_map[*vi]);
      }
    }
  }
  offset = predecessors_ready - current_time;
  //advance in time profile
  if(offset > 0) { 
    advance(offset); 
  }
	
  //latency of the instruction, length of reservation table
  int latency = 0;
  int my_value_ready = 0;/* calculate by finding biggest reserv_tab sub array */
  for(int i = 0; i < _reserv_tab.size(); i++){
    my_value_ready = max(my_value_ready, current_time + _reserv_tab[i].size());
  }

  /* Advance until insertable */
 _recheck:	
  for(int i = 0; i < _reserv_tab.size(); i++) {
    latency = max(latency, _reserv_tab[i].size());
    int check_size = min((*rtab)[i].size(), _reserv_tab[i].size());
		
    for(int j = 0; j < check_size; j++) {
      if((_reserv_tab[i][j] == BUSY) && ((*rtab)[i][j] == BUSY)) {
	advance(1);
	offset++;
	my_value_ready++;
	goto _recheck;
      }
    }
  }
  /* Insert my value into the ready map */
  for(vertex_set_t::iterator nit=_nodes.begin(); nit!=_nodes.end(); ++nit){
    ready_map[*nit] = my_value_ready;
  }

  //We can insert now. The latency of the instruction is latency.
  insert_reserv_tab(_nodes, _reserv_tab, latency);
  //Comput hash_value
  compute_hash();
	
  return offset;
}

int TimeProfile::get_depth(void)
{
  int depth = 0, size = (*rtab).size();
  for(int i = 0; i < size; i++) {
    depth = max(depth, ((*rtab)[i]).size());
  }
  return depth;
}

///Advance a given number of clock cycles
void TimeProfile::advance(int _offset)
{
  current_time+=_offset;
  for(int i = 0; i < (*rtab).size(); i++) {
    for(int j = 0; j < _offset; j++) {
      if((*rtab)[i].size()) {
	(*rtab)[i].erase((*rtab)[i].begin());
      }
    } 
  }
  //Update nodes in the TimeProfile

  //mater: I think this had a nasty bug in it, erasing the item an
  //iterator points to could mean problems.
  vertex_set_t::iterator tmp;
  for(vertex_set_t::iterator vi = nodes.begin();
      vi != nodes.end(); ) {
    tmp = vi;
    tmp++;
    if((latency_map[*vi] -= _offset) <= 0) {
      latency_map.erase(*vi);
      nodes.erase(vi); 
    }
    vi = tmp;
  }
}

void TimeProfile::insert_reserv_tab(const vertex_set_t & _nodes,
				    const reserv_tab_t & _reserv_tab, int _latency)
{
  for(int i = 0; i < _reserv_tab.size(); i++) {
    int tp_entry_size = (*rtab)[i].size();
    int rt_entry_size = _reserv_tab[i].size();
    if(tp_entry_size < rt_entry_size) {
      //Need new entries in TimeProfile vector
      for(int j = 0; j < tp_entry_size; j++) {
	if(_reserv_tab[i][j] == BUSY) {
	  assert((*rtab)[i][j] == EMPTY);
	  (*rtab)[i][j] = BUSY;
	}
      }
      for(int j = tp_entry_size; j < rt_entry_size; j++) {
	(*rtab)[i].push_back(_reserv_tab[i][j]);
      }
      assert((*rtab)[i].size() == rt_entry_size);
    } else {
      for(int j = 0; j < rt_entry_size; j++) {
	if(_reserv_tab[i][j] == BUSY) {
	  assert((*rtab)[i][j] == EMPTY);
	  (*rtab)[i][j] = BUSY;
	}
      }
    }
  }
  //Insert nodes and their latency (each node in the pattern has the same
  //latency).
  for(vertex_set_t::const_iterator vi = _nodes.begin(); vi != _nodes.end();
      ++vi) {
    if(_latency > 0) {
      assert(nodes.find(*vi) == nodes.end());
      assert(latency_map.find(*vi) == latency_map.end());
      nodes.insert(*vi);
      latency_map[*vi] = _latency;
    }
  }
}

void TimeProfile::compute_hash(void)
{
  hash_value = 0;

  //Time profile shape
  for(int i = 0; i < (*rtab).size(); i++) {
    for(int j = 0; j < (*rtab)[i].size(); j++) {
      hash_value += primes[i % MP][j % MP];
    }
  }
  //Nodes
  size_t nodes_hash = 1;
  for(vertex_set_t::const_iterator vi = nodes.begin(); vi != nodes.end();
      ++vi) {
    nodes_hash *= (size_t)(*vi);
  }
  hash_value += nodes_hash;
}

int TimeProfile::get_ready(vertex_id v) const
{
  if(ready_map.find(v) != ready_map.end())
    return (*(ready_map.find(v))).second;
  std::cerr << "Warning, checking " << v << " when it is not in the ready_map. " << std::endl;
  return 1000; /* Bogus value */
}

//Debug
std::ostream & operator<<(std::ostream & _out, const TimeProfile & _time_profile) {
  int f = _time_profile.size();
  _out << _time_profile.current_time << std::endl;
  _out << "nodes: { ";
  std::copy(_time_profile.nodes.begin(), _time_profile.nodes.end(), 
      std::ostream_iterator< vertex_id >(_out, " "));
  _out << "}\n";
  _out << "latencies: (v, l, r): { ";

  std::map< vertex_id, int > latencies = _time_profile.latency_map;
  std::map< vertex_id, int >::iterator li = latencies.begin(),
    li_end = latencies.end();
  for(; li != li_end; ++li) {
    _out << "(" << (*li).first << ", " << (*li).second << ", " 
	 << _time_profile.get_ready((*li).first) << ") ";
  }

  _out << "}\ntime profile:\n";
  for(int i = 0; i < f; i++) {
    int size = _time_profile.get_rtab()[i].size();
    _out << i << " (" << size << "): ";
    for(int j = 0; j < size; j++) {
      _out << _time_profile.get_rtab()[i][j] << " ";
    }
    _out << std::endl;
  }
  return _out;
}
